
import numpy
import itertools
import time
from traits.api import Instance, Title, Float, Tuple, Complex, Property, cached_property, observe, Bool, Range
from traitsui.api import View, Item, VGroup, Tabbed, Include, Group
from tvtk.api import tvtk

from raypier.sources import BaseRaySource, UnitTupleVector, UnitVectorTrait
from raypier.core.ctracer import GaussletCollection, gausslet_dtype, ray_dtype
from raypier.tracer import normaliseVector
from raypier.editors import NumEditor
from raypier.decompositions import decompose_angle, make_hexagonal_grid


class BaseGaussletSource(BaseRaySource):
    """
    Abstract base class for Gausslet Sources.
    """
    #:object title
    name = Title("Gausslet Source", allow_selection=False)
    
    para_data_source = Instance(tvtk.ProgrammableSource, transient=True)
    para_tube = Instance(tvtk.TubeFilter, (), transient=True)
    para_ray_actor = Instance(tvtk.Actor, transient=True)
    para_property = Instance(tvtk.Property, (), {'color':(0.0,0.5,0.5), 'opacity': 1.0}, transient=True)
    para_mapper = Instance(tvtk.PolyDataMapper, (), transient=True)
    
    #: set True to render the parabasal rays
    show_paras = Bool(False)
    
    #: wavelength, in microns.
    wavelength = Float(0.78) # in microns
    
    #: Total power in the source beam, in Watts, calculated as the sum of the power in each individual
    #: gausslet. This is, in fact,  *not* the true beam power. Due to the overlap between the modes,
    #: the actual beam power due to the coherent sum of the modes is greater than the sum of the 
    #: single-mode powers.
    beam_power = Float(1.0)
    
    display_grp = VGroup(Item('display'),
                       Item('show_start'),
                       Item('show_normals'),
                       Item('show_paras'),
                       Item('scale_factor'),
                       Item('export_pipes',label="export as pipes"),
                       Item('opacity', editor=NumEditor),
                       label="Display")
    
    params_grp = VGroup(
                       Item("max_angle"),
                       Item('beam_waist', tooltip="1/e^2 beam intensity radius, in microns"),
                       Item('sample_spacing', tooltip="Sample spacing for the field in the aperture"),
                       label="Parameters")
    
    traits_view = View(Item('name', show_label=False),
                       Tabbed(VGroup(Group(Item('origin', show_label=False,resizable=True), 
                                           show_border=True,
                                    label="Origin position",
                                    padding=0),
                               Group(Item('direction', show_label=False, resizable=True),
                                    show_border=True,
                                    label="Direction"),
                               Item("wavelength"),
                               Include('params_grp'),
                            ),
                              display_grp)
                       )
    
    def _traced_rays_changed(self):
        self.data_source.modified()
        self.para_data_source.modified()
        self.normals_source.modified()
        self._mtime = time.monotonic()
        
    def _input_rays_changed(self):
        self.update=True
        
    def _opacity_changed(self):
        self.vtkproperty.opacity = self.opacity
        self.para_property.opacity = self.opacity
        self.render = True
        
    def _get_actors(self):
        actors = [self.ray_actor, self.para_ray_actor, self.start_actor, self.normals_actor]
        return actors
    
    def _normals_source_default(self):
        source = tvtk.ProgrammableSource()
        
        def get_origins(gc_list):
            for gc in gc_list:
                yield gc.base_rays.origin.reshape(-1,3)
                if self.show_paras:
                    yield gc.para_origin.reshape(-1,3)
                
        def get_normals(gc_list):
            for gc in gc_list:
                yield gc.base_rays.normal.reshape(-1,3)
                if self.show_paras:
                    yield gc.para_normal.reshape(-1,3)
                
        def execute():
            if not self.show_normals:
                return
            output = source.poly_data_output
            points = numpy.vstack( list( get_origins(self.traced_rays[1:]) ) )
            normals = numpy.vstack( list( get_normals(self.traced_rays[1:]) ) )
            output.points = points
            output.point_data.normals = normals
            #print("calc normals GLYPH")
        source.set_execute_method(execute)
        return source
    
    @observe("show_paras")
    def _on_show_paras_changed(self, evt):
        self.para_data_source.modified()
        self.render=True
    
    def _para_data_source_default(self):
        source = tvtk.ProgrammableSource()
        def execute():
            if not self.show_paras:
                return
            output = source.poly_data_output
            pointArrayList = []
            mask = self.ray_mask
            for rays in self.traced_rays:
                vis = mask.get(rays, [])
                if len(vis) != len(rays):
                    vis = itertools.repeat(True)
                start_pos_ = (r.parabasal_rays for r,v in zip(rays, vis) if v)
                start_pos = [p.origin for p in itertools.chain.from_iterable(start_pos_)]
                end_pos_ = (r.parabasal_rays for r,v in zip(rays, vis) if v)
                end_pos = [p.termination for p in itertools.chain.from_iterable(end_pos_)]
                interleaved = numpy.array([start_pos, end_pos]).swapaxes(0,1).reshape(-1,3)
                pointArrayList.append(interleaved)
            if pointArrayList:
                points = numpy.vstack(pointArrayList)
                output.points=points
                lines = list([i,i+1] for i in range(0,points.shape[0],2))
                output.lines = lines
        source.set_execute_method(execute)
        return source
    
    def _para_ray_actor_default(self):
        tube = self.para_tube
        tube.radius = self.scale_factor
        tube.input_connection=self.para_data_source.output_port
        tube.number_of_sides = 20
        
        map = self.para_mapper
        act = tvtk.Actor(mapper=map)
        act.visibility = True
        display = self.display
        if display=="pipes":
            map.input_connection=tube.output_port
        elif display=="wires":
            map.input_connection = self.para_data_source.output_port
        else:
            act.visibility = False
        prop = self.para_property
        prop.opacity = self.opacity
        if prop:
            act.property = prop
        return act
    
    def _scale_factor_changed(self, scale):
        self.tube.radius = scale
        self.para_tube.radius = scale
        self.normals_glyph.scale_factor=self.scale_factor*10
        self.sphere.radius = self.scale_factor*3
        self.render = True
        
    def _display_changed(self, vnew):
        actors = self.actors
        for act in actors:
            act.visibility = True
        if vnew=="pipes":
            self.mapper.input_connection = self.tube.output_port
            self.para_mapper.input_connection = self.para_tube.output_port
        elif vnew=="wires":
            self.mapper.input_connection = self.data_source.output_port
            self.para_mapper.input_connection = self.para_data_source.output_port
        elif vnew=="hidden":
            for act in actors:
                act.visibility = False
        self.render = True
        
        
class SingleGaussletSource(BaseGaussletSource):
    """
    A source which launches a single Gausslet. Mainly used for testing.
    """
    abstract=False
    
    #: The origin of the Gausslet base-ray
    origin = Tuple((0.,0.,0.))
    
    #: Base-ray direction unit-vector.
    direction = UnitTupleVector
    
    #: Polarisation unit-vector for the E1 amplitude. Orthogonal to the direction vector.
    E_vector = UnitVectorTrait((1.,0.,0.), editor_traits={'cols':3,
                                'labels':['x','y','z']})
    
    #: Complex amplitude for the E1_vector polarisation direction. 
    E1_amp = Complex(1.0+0.0j)
    
    #: Complex amplitude for the E2_vector, implicitly orthogonal to both `direction` and `E1_vector`.
    E2_amp = Complex(0.0+0.0j)
    
    #: Gaussian 1/e**2 beam radius, in microns, at the beam waist.
    beam_waist = Float(1.0) #radius, in microns
    
    #: The GaussletCollection launched by this object.
    input_rays = Property(Instance(GaussletCollection), 
                         depends_on="origin, direction, max_ray_len, E_vector, E1_amp, E2_amp, beam_waist")
    
    params_grp = VGroup(
                       Item("wavelength"),
                       Item("max_angle"),
                       Item('beam_waist', tooltip="1/e^2 beam intensity radius, in microns"),
                       Item('sample_spacing', tooltip="Sample spacing for the field in the aperture"),
                       label="Parameters")
    
    def _wavelength_list_default(self):
        return [self.wavelength,]
    
    @observe("wavelength")
    def _do_wavelength_changed(self, evt):
        self.wavelength_list = [self.wavelength]
        self.update=True
    
    @cached_property
    def _get_input_rays(self):
        origin = numpy.array(self.origin)
        direction = numpy.array(self.direction)
        max_axis = numpy.abs(direction).argmax()
        if max_axis==0:
            v = numpy.array([0.,1.,0.])
        else:
            v = numpy.array([1.,0.,0.])
        d1 = numpy.cross(direction, v)
        d1 = normaliseVector(d1)
        d2 = numpy.cross(direction, d1)
        d2 = normaliseVector(d2)

        E_vector = numpy.cross(direction, self.E_vector)
        E_vector = numpy.cross(E_vector, direction)
        
        angles = (numpy.arange(6)*numpy.pi/3.)[:,None]
        offsets = numpy.cos(angles)*d1[None,:] + numpy.sin(angles)*d2[None,:]
        offsets *= self.beam_waist
        
        data = numpy.zeros((1,), dtype=gausslet_dtype)
        base_ray = data['base_ray']
        base_ray['origin'] = origin[None,:]
        base_ray['direction'] = direction[None,:]
        base_ray['E_vector'] = E_vector[None,:]
        base_ray['E1_amp'] = 1.0
        base_ray['length'] = self.max_ray_len
        
        para = data['para_rays']
        para['direction'] = direction[None,None,:]
        para['origin'] = offsets[None,:,:]
        para['length'] = self.max_ray_len
        para['normal'] = -direction[None,:]
        
        gc = GaussletCollection.from_array(data)
        #print("A", [g.length for g in gc[0].parabasal_rays] , [self.max_ray_len]*6 )
        assert [g.length for g in gc[0].parabasal_rays] == [self.max_ray_len]*6
        return gc
    
    
class TopHatGaussletSource(SingleGaussletSource):
    """
    Launches Gausslets forming a top-hat intensity profile at the beam-waist.
    """
    
    #: The diameter of the beam at the beam-waist, in microns.
    beam_waist = Float(100.0) #diameter, in microns
    
    #: The sample-spacing across the beam waist. 
    sample_spacing = Float(10.0) #in microns
    
    #: The maximum divergence angle for the Gausslet base-rays.
    max_angle = Float(10.0)
    
    input_rays = Property(Instance(GaussletCollection), 
                         depends_on="origin, direction, beam_waist, wavelength, "
                         "max_ray_len, E_vector, sample_spacing, max_angle")
    
    geom_grp = VGroup(Group(Item('origin', show_label=False,resizable=True), 
                            show_border=True,
                            label="Origin position",
                            padding=0),
                       Group(Item('direction', show_label=False, resizable=True),
                            show_border=True,
                            label="Direction"),
                       Item("wavelength"),
                       Item("max_angle"),
                       Item('beam_waist', tooltip="1/e^2 beam intensity radius, in microns"),
                       Item('sample_spacing', tooltip="Sample spacing for the field in the aperture"),
                       label="Geometry")
    
    
    @observe("direction, beam_waist, max_ray_len, sample_spacing, max_angle")
    def on_update(self, evt):
        self.data_source.modified()
        self.update=True
    
    @cached_property
    def _get_input_rays(self):
        try:
            origin = numpy.array(self.origin)
            direction = numpy.array(self.direction)
            radius = self.beam_waist/2.0 #convert to mm
            
            max_axis = numpy.abs(direction).argmax()
            if max_axis==0:
                v = numpy.array([0.,1.,0.])
            else:
                v = numpy.array([1.,0.,0.])
                
            max_angle=self.max_angle
            input_spacing = self.sample_spacing
            wavelength = self.wavelength
            oversample=4
            
            x = numpy.linspace(-radius,radius, int(radius/input_spacing)+1)
            X,Y = numpy.meshgrid(x,x)
            select = (X**2 + Y**2) < radius
            E_field = numpy.zeros((2,len(x),len(x)), dtype=numpy.complex128)
            E1 = E_field[0,:,:]
            E1[select] = 1.0 
                
            rays = decompose_angle(origin, direction, v, E_field, input_spacing, max_angle, wavelength, oversample)
                
        except:
            import traceback
            traceback.print_exc()
            return GaussletCollection(5)
        return rays
    
    
class CollimatedGaussletSource(SingleGaussletSource):
    """
    Launches a quasi-collimated beam of Gausslets.
    """
    
    #: The max radius of the Gausslets at the objects origin.
    radius = Float(10.,editor=NumEditor)
    
    #: Sets the number of Gausslets found along the radius of the beam, 
    #: I.e. Controls the spacing of the Gausslets, being radius/resolution.  
    resolution = Float(10.0, editor=NumEditor)
    
    #: Controls the overlap between adjacent Gausslets. If blending==1.0, the 
    #: Gaussian 1/e^2 intensity width is equal to the Gausslet spacing.
    #: The Gausslet width = spacing/blending. I.e. a higher value reduces the Gausslet overlap.
    blending = Float(1.5)
    
    input_rays = Property(Instance(GaussletCollection), 
                         depends_on="origin, direction, max_ray_len, E_vector, E1_amp, "
                                    "E2_amp, radius, resolution, wavelength, beam_waist, blending")
    
    params_grp = VGroup(
                        Item("radius", editor=NumEditor),
                        Item("resolution", editor=NumEditor),
                       Item('beam_waist', tooltip="1/e^2 beam intensity radius, in microns", editor=NumEditor),
                       Item("blending", editor=NumEditor),
                       label="Parameters")
    
    @cached_property
    def _get_input_rays(self):
        self.wavelength_list = [self.wavelength]
        origin = numpy.array(self.origin)
        direction = numpy.array(self.direction)
        radius = self.radius
        spacing = radius/self.resolution
        max_axis = numpy.abs(direction).argmax()
        if max_axis==0:
            v = numpy.array([0.,1.,0.])
        else:
            v = numpy.array([1.,0.,0.])
        d1 = numpy.cross(direction, v)
        d1 = normaliseVector(d1)
        d2 = numpy.cross(direction, d1)
        d2 = normaliseVector(d2)

        E_vector = numpy.cross(self.E_vector, direction)
        E_vector = numpy.cross(E_vector, direction)
        
        cosV = numpy.cos(numpy.pi*30/180.)
        sinV = numpy.sin(numpy.pi*30/180.)
        d2 = cosV*d2 + sinV*d1
        
        nsteps = int(radius/(spacing*numpy.cos(numpy.pi*30/180.)))
        i = numpy.arange(-nsteps-1, nsteps+2)
        j = numpy.arange(-nsteps-1, nsteps+2)
        
        vi, vj = numpy.meshgrid(i,j)
        
        vi.shape = -1
        vj.shape = -1
        
        ri = ((vi + sinV*vj)**2 + (cosV*vj**2))
        select = ri < (radius/spacing)**2
        
        xi = vi[select]
        yj = vj[select]
        
        offsets = xi[:,None]*d1 + yj[:,None]*d2
        offsets *= spacing
        gauss = numpy.exp(-(ri[select]*(spacing**2)/((self.beam_waist)**2))) 
        
        ray_data = numpy.zeros(offsets.shape[0], dtype=ray_dtype)
        
        E1_amp = self.E1_amp
        E2_amp = self.E2_amp
        #print("E_vals", E1_amp, E2_amp)
        P1 = (E1_amp * E1_amp.conjugate()).real
        P2 = (E2_amp * E1_amp.conjugate()).real
        
        ### The true E_field is E1_amp (or E2_amp) * sqrt(area of beam)###
        beamlet_area = (spacing**2) * numpy.pi
        total_power = ((P1 + P2) * (gauss**2)).sum() #* beamlet_area
        scaling = numpy.sqrt(self.beam_power/total_power)
            
        ray_data['origin'] = offsets
        ray_data['origin'] += origin
        ray_data['direction'] = direction
        ray_data['wavelength_idx'] = 0
        ray_data['E_vector'] = [normaliseVector(E_vector)]
        ray_data['E1_amp'] = scaling * self.E1_amp * gauss
        ray_data['E2_amp'] = scaling * self.E2_amp * gauss
        ray_data['refractive_index'] = 1.0+0.0j
        ray_data['normal'] = [[0,1,0]]
        ray_data['ray_type_id'] = (1<<1) #indicates Gausslets
        
        rays = GaussletCollection.from_rays(ray_data)
        wl = numpy.array(self.wavelength_list)
        rays.wavelengths = wl
        working_dist = 0.0
        rays.config_parabasal_rays(wl, spacing/self.blending, working_dist)
        return rays
        
        
class GaussianPointSource(SingleGaussletSource):
    """
    Creates a low-F# point source with symmetric Gaussian profile.
    
        resolution - the number of rays along the angular radius.
        
        working_dist - the distance in front of the origin plane at which the focus is found
        
        numerical_aperture - equal of the sin of the maximum ray angle
        
        blending - determines the mode overlap
    """
    abstract = False
    
    #: Number of angle steps along the radius
    resolution = Range(0.0, None, 10.0)
    
    #: The distance from the object origin to the beam waist. A negative value
    #: means that the focus is behind the origin. 
    working_dist = Float(100.0)
    
    #: The maximum numerical aperture of the base-rays in the beam.
    numerical_aperture = Float(0.2)
    
    #: Sets the overlap between adjacent Gausslets *in the far field*.
    #: I.e. a higher value increases the Gausslet beam waist, which
    #: reduces the divergence of the Gaussian modes such that their
    #: overlap in the far field is reduced.
    blending = Float(1.0)
    
    ### Traits Properties seems to be broken. No good way to flush the cache so I'm doing it myself.
    _input_rays = Instance(GaussletCollection)
    input_rays = Property(Instance(GaussletCollection))
    
    params_grp = VGroup(
                        Item("numerical_aperture", editor=NumEditor),
                        Item("resolution", editor=NumEditor),
                       Item('beam_waist', tooltip="1/e^2 beam intensity radius, in microns", editor=NumEditor),
                       Item('working_dist', editor=NumEditor),
                       Item("blending", editor=NumEditor),
                       label="Parameters")
    
    @observe("origin, direction, max_ray_len, E_vector, E1_amp, "\
            "E2_amp, numerical_aperture, resolution, wavelength, "\
            "beam_waist, blending, working_dist")
    def _clear_input_rays(self, evt):
        del self._input_rays
        self.update = True
    
    def _get_input_rays(self):
        rays = self._input_rays
        if rays is None:
            rays = self._eval_input_rays()
            self._input_rays = rays
        return rays
    
    def _eval_input_rays(self):
        self.wavelength_list = [self.wavelength]
        origin = numpy.array(self.origin)
        direction = normaliseVector(numpy.array(self.direction))
        
        d2 = normaliseVector(numpy.cross(direction, self.E_vector))
        d1 = numpy.cross(direction, d2)
        
        focal_radius = self.beam_waist/1000.0 #convert to mm
        
        n = 1.0 + 0.0j #refractive index of medium
        k_abs = (2000.0*numpy.pi)/self.wavelength #k now in inverse-mm
        k_radius = 2./focal_radius
        k_max = self.numerical_aperture * k_abs
        k_grid_spacing = k_max / self.resolution
        
        kx,ky = make_hexagonal_grid(k_max, spacing=k_grid_spacing)
        
        krsq = kx**2 + ky**2
        kz = numpy.sqrt(k_abs**2 - krsq)
        
        waist_location = origin + self.working_dist*direction 
        directions = (kx[:,None]*d1 + ky[:,None]*d2 + kz[:,None]*direction)/k_abs
        directions = normaliseVector(directions)
        
        gauss = numpy.exp(-(krsq/(k_radius**2)))
        E1_amp = gauss * self.E1_amp
        E2_amp = gauss * self.E2_amp
        
        ray_data = numpy.zeros(gauss.shape[0], dtype=ray_dtype)
            
        E_vectors = numpy.cross(directions, d2[None,:])
            
        ray_data['origin'] = waist_location[None,:]
        ray_data['direction'] = directions
        ray_data['wavelength_idx'] = 0
        ray_data['E_vector'] = E_vectors
        ray_data['E1_amp'] = E1_amp
        ray_data['E2_amp'] = E2_amp
        ray_data['refractive_index'] = 1.0+0.0j
        ray_data['normal'] = [[0,1,0]]
        ray_data['ray_type_id'] = (1<<1) #indicates Gausslets
        
        rays = GaussletCollection.from_rays(ray_data)
        wl = numpy.array(self.wavelength_list)
        rays.wavelengths = wl
        working_dist = 0.0
        gauss_radius = 2.*self.blending/(k_grid_spacing)
        rays.config_parabasal_rays(wl, gauss_radius, working_dist)
        rays.project_to_plane(origin, direction)
        return rays
    



from traits.api import Float, Instance, Bool, on_trait_change, Array, Enum
from traitsui.api import View, Item, VGroup, HGroup, EnumEditor
from chaco.api import GridDataSource, GridMapper, ImageData, Spectral,\
        DataRange1D, CMapImagePlot, DataRange2D, PlotComponent, Plot,\
        ArrayPlotData, PlotComponent, HPlotContainer, ColorBar, LinearMapper, ImagePlot
        
from chaco.default_colormaps import Spectral
        
from enable.api import ComponentEditor, KeySpec
from chaco.tools.api import PanTool, ZoomTool

from .results import Result
from .fields import EFieldPlane

import numpy
from numpy import inf


class ProbePlanePanTool(PanTool):
    probe = Instance(EFieldPlane)
    
    def normal_mouse_wheel(self, event):
        print(event)
        shift = event.mouse_wheel_delta[1]/120.0
        probe = self.probe
        span = min(probe.width, probe.height)
        delta = shift * span/100.0
        
        probe.translate_local(0, 0, delta)
    
    def panning_mouse_move(self, event):
        """ Handles the mouse being moved when the tool is in the 'panning'
        state.
        """
        plot = self.component

        if self._auto_constrain and self.constrain_direction is None:
            # Determine the constraint direction
            x_orig, y_orig = self._original_xy
            if abs(event.x - x_orig) > abs(event.y - y_orig):
                self.constrain_direction = "x"
            else:
                self.constrain_direction = "y"

        deltas = [0.0,0.0,0.0]
        direction_info = [("x", "width", 0), ("y", "height", 1)]
        for direction, bound_name, index in direction_info:
            if not self.constrain or self.constrain_direction == direction:
                mapper = getattr(plot, direction + "_mapper")
                domain_min, domain_max = mapper.domain_limits
                eventpos = getattr(event, direction)
                origpos = self._original_xy[index]

                deltas[index] = mapper.map_data(origpos) - mapper.map_data(eventpos)

                # Use .set_bounds() so that we don't generate two range_changed
                # events on the DataRange
        if event.control_down:
            #do zoom
            self.probe.width += deltas[0]
            self.probe.height += deltas[1] 
        else:
            self.probe.translate_local(deltas[0], deltas[1], deltas[2])

        event.handled = True

        self._original_xy = (event.x, event.y)
        #plot.request_redraw()
        return


class IntensityImageView(Result):
    name = "Field Plane Image View"
    field_probe = Instance(EFieldPlane)
    
    display = Enum("Intensity", "E_x", "E_y", "E_z")
    use_log = Bool(False)
    
    intensity_data = Array()
    
    hbox = Instance(HPlotContainer)
    cbar = Instance(ColorBar)
    crange = Instance(DataRange1D, ())
    
    ds = Instance(ArrayPlotData, ())
    plot = Instance(Plot)
    image_plot = Instance(ImagePlot)
    
    pan_tool = Instance(PanTool)
    
    width = Float(1.0)
    height = Float(1.0)
    
    traits_view = View(VGroup(
            HGroup(Item("display", style="simple", show_label=False),
                   Item("use_log", show_label=True)),
            Item("hbox", editor=ComponentEditor(), show_label=False)
                ))
    
    
    def calc_intensity(self, e_field):
        E = e_field
        mode = self.display
        if mode=="Intensity":
            U = self.field_probe.intensity
        else:
            idx = {"E_x":0, "E_y":1, "E_z":2}[mode]
            U = E[:,:,idx]
            
        if self.use_log:
            U = numpy.log10(U)
            
        self.intensity_data = U
        return U
    
    @on_trait_change("use_log, display")
    def _display_changed(self):
        self.on_field_changed(self.field_probe.E_field)
    
    def calc_result(self, model):
        pass
        #self.calc_intensity(self.field_probe.E_field)
        
    def _pan_tool_default(self):
        #return PanTool()
        return ProbePlanePanTool()
    
    def _field_probe_changed(self, old, new):
        if old is not None:
            old.on_trait_change(self.on_field_changed, "E_field", remove=True)
        new.on_trait_change(self.on_field_changed, "E_field")
        self.pan_tool.probe = new
        
    def on_field_changed(self, e_field):
        intensity_image = self.calc_intensity(e_field)
        
        probe = self.field_probe
        self.ds.set_data("img", intensity_image)
        side = probe.width
        yside = probe.height
        
        if self.plot is not None:
            a = probe.centre
            self.plot.title = f"Intensity @ ({a[0]:0.3f}, {a[1]:0.3f}, {a[2]:0.3f})"
            
            plot = self.image_plot
            #plot.
            plot.x_mapper.range.set_bounds(0,side)
            plot.y_mapper.range.set_bounds(0,yside)
            xdata = numpy.linspace(0,side,probe.size)
            ydata = numpy.linspace(0,yside, probe.size)
            plot.index.set_data(xdata,ydata)
            plot.request_redraw()
            
    def _hbox_default(self):
        hbox = HPlotContainer()
        hbox.add(self.plot)
        hbox.add(self.cbar)
        #hbox.bgcolor = 'sys_window' 
        return hbox
    
    def _cbar_default(self):
        colormap = Spectral(self.crange) #self.cmap(self.crange)
        cbar = ColorBar(index_mapper=LinearMapper(range=self.crange),
                        color_mapper=colormap,
                        padding=20, width=20, orientation='v', resizable='v')
        axis = cbar._axis
        #axis.tick_label_formatter = tick_formatter
        #cbtool = ColorbarTool(component=cbar, color_map=self.cmap.__name__)
        #cbar.tools.append(cbtool)
        self.crange.on_trait_change(self.on_rescale, "updated")
        return cbar
    
    def on_rescale(self):
        if self.hbox is not None:
            self.hbox.request_redraw()
    
    def _plot_default(self):
        probe = self.field_probe
        if probe is None:
            xside=1
            yside=1
            data = [[0.0]]
        else:
            xside = probe.width
            yside = probe.height
            data = self.calc_intensity(probe.E_field)
            
        x_bounds = (0,xside)
        y_bounds = (0,yside)
        
        ds = self.ds
        ds.set_data("img", self.intensity_data)
        plot = Plot(ds)
        imgplot = plot.img_plot("img",
                                xbounds=x_bounds,
                                ybounds=y_bounds,
                                colormap=Spectral(self.crange))[0]
        self.image_plot = imgplot
        self.cbar.index_mapper.range.sources = [self.image_plot.value]
        plot.x_axis.title = "Width /mm"
        plot.y_axis.title = "Height /mm"
        a = probe.centre
        plot.title = f"Intensity @ ({a[0]:0.3f}, {a[1]:0.3f}, {a[2]:0.3f})"
        
        pantool = PanTool(plot) #self.pan_tool
        pantool = self.pan_tool
        pantool.component = self.image_plot
        self.pan_tool = pantool
        plot.tools.append(pantool)
        
        return plot



#    Copyright 2009, Teraview Ltd.
#
#    This file is part of Raypier.
#
#    Raypier is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



from traits.api import HasTraits, Array, Float, Complex,\
            Property, List, Instance, Range, Any,\
            Tuple, Event, cached_property, Set, Int, Trait, Button,\
            self, Str, Bool, PythonValue, Enum, File
from traitsui.api import View, Item, ListEditor, VSplit,\
            RangeEditor, ScrubberEditor, HSplit, VGroup, TextEditor,\
            TupleEditor, VGroup, HGroup, TreeEditor, TreeNode, TitleEditor,\
            ShellEditor, Controller, Tabbed
            
from traitsui.menu import Menu, MenuBar, Action, Separator
            
from pyface.api import FileDialog, OK as FD_OK
            
from tvtk.api import tvtk
from tvtk.pyface.scene_model import SceneModel
from tvtk.pyface.scene_editor import SceneEditor
import numpy
import threading, os, itertools
import os
import traceback
import yaml
from contextlib import contextmanager
from itertools import chain, islice, count
from raypier.sources import BaseRaySource
from raypier.core.ctracer import Face, RayCollection
from raypier.core.tracer import trace_rays
from raypier.constraints import BaseConstraint
from raypier.has_queue import HasQueue, on_trait_change
from raypier.bases import Traceable, Probe, Result
from raypier.utils import normaliseVector, transformNormals, transformPoints,\
        transformVectors, dotprod
        
from raypier.core import ctracer
from raypier.qt_future_call import FutureCall

counter = count()

from raypier import mirrors, prisms, corner_cubes, ellipsoids, sources,\
    results, beamstop, lenses, beamsplitters, waveplates, apertures, fields, gausslet_sources

optics_classes = sorted(Traceable.subclasses, key=lambda c: c.__name__)

optics_menu = Menu(name = "Components...",
                   *[Action(name=cls.__name__, action="insert_"+cls.__name__)\
                    for cls in optics_classes]
                    )

source_classes = sorted(BaseRaySource.subclasses, key=lambda c:c.__name__)
sources_menu = Menu(name = "Sources...",
                   *[Action(name=cls.__name__, action="insert_"+cls.__name__)\
                    for cls in source_classes]
                    )

results_classes = sorted(Result.subclasses, key=lambda c:c.__name__)
results_menu = Menu(name = "Results...",
                   *[Action(name=cls.__name__, action="insert_"+cls.__name__)\
                    for cls in results_classes]
                    )

menubar = MenuBar(Menu(Action(name="Open...", action="open_file_action"),
                       Action(name="Save...", action="save_file_action"),
                       Action(name="Save As...", action="save_as_action"),
                       name="File..."),
                  Menu(optics_menu,
                       sources_menu,
                       results_menu,
                       name="Insert..."))
    
    
class RayTraceModelHandler(Controller):
    def save_file_action(self, info):
        model = info.ui.context['object']
        try:
            model.save_as_yaml()
        except IOError:
            self.save_as_action(info)
        
    def save_as_action(self, info):
        dlg = FileDialog(parent=info.ui.control,
                         action='save as', 
                         wildcard=FileDialog.create_wildcard("Raypier model", ["*.raymod"]))
        ret = dlg.open()
        if ret == FD_OK:
            fname = dlg.path
            model = info.ui.context['object']
            model.save_as_yaml(filename=fname)
        info.ui.title = fname
    
    def open_file_action(self, info):
        dlg = FileDialog(parent=info.ui.control,
                         action='open', 
                         wildcard=FileDialog.create_wildcard("Raypier model", ["*.raymod"]))
        ret = dlg.open()
        if ret == FD_OK:
            fname = dlg.path
            if os.path.exists(fname):
                model = info.ui.context['object']
                model.load_from_yaml(fname)
            info.ui.title = fname
        
    def insert_component(self, info, cls):
        tracer = info.ui.context['object']
        tracer.optics.append(cls())
        
    def insert_source(self, info, cls):
        tracer = info.ui.context['object']
        tracer.sources.append(cls())
        
    def insert_result(self, info, cls):
        tracer = info.ui.context['object']
        tracer.results.append(cls())
        

for cls in optics_classes:
    def make_action(_cls):
        def action(self, info):
            self.insert_component(info, _cls)
        return action
    setattr(RayTraceModelHandler, "insert_"+cls.__name__, make_action(cls))
    
for cls in source_classes:
    def make_action(_cls):
        def action(self, info):
            self.insert_source(info, _cls)
        return action
    setattr(RayTraceModelHandler, "insert_"+cls.__name__, make_action(cls))
    
for cls in results_classes:
    def make_action(_cls):
        def action(self, info):
            self.insert_result(info, _cls)
        return action
    setattr(RayTraceModelHandler, "insert_"+cls.__name__, make_action(cls))
    
    
class RayTraceModel(HasQueue):
    scene = Instance(SceneModel, (), {'background':(1,1,0.8)}, transient=True)
    
    optics = List(Traceable)
    sources = List(BaseRaySource)
    probes = List(Probe)
    constraints = List(BaseConstraint)
    results = List(Result)
    
    optical_path = Float(0.0, transient=True)
    
    all_faces = List(ctracer.Face, desc="global list of all faces, created automatically "
                     " when a tracing operation is initiated. Ray end_face_idx "
                     "can be used to index this list")
    face_sets = List(ctracer.FaceList, desc="list of FaceLists extracted from all "
                     "optics when a tracing operation is initiated")
    
    update = Event() #triggers a tracing operation
    _updating = Bool(False) #indicating that tracing is in progress
    update_complete = Event()
    
    _hold_off = Bool(False) #Blocks tracing (while model parameters are altered)
    _update_requested = Bool(False)
    
    Self = self
    ShellObj = PythonValue({}, transient=True)
        
    recursion_limit = Int(200, desc="maximum number of refractions or reflections")
    
    save_btn = Button("Save scene")
    
    filename = File()
    
    def load_from_yaml(self, filename):
        with open(filename, 'r') as fobj:
            model = yaml.load(fobj)
        print(model)
        self.optics = model['components']
        self.sources = model['sources']
        self.results = model['results']
        self.update = True
    
    def save_as_yaml(self, filename=None):
        if filename is None:
            filename = self.filename
            if self.filename is None:
                raise IOError("no preset filename")
        model = {"components":list(self.optics),
                 "sources": list(self.sources),
                 "results": list(self.results)}
        with open(filename, 'w') as fobj:
            yaml.dump(model, fobj)
        self.filename = filename
    
    @on_trait_change("optics[]")
    def on_optics_change(self, obj, name, removed, opticList):
        #print("adding", opticList, removed, name)
        scene = self.scene
        #del scene.actor_list[:]    
        for o in opticList:
            scene.add_actors(o.get_actors(scene))
        for o in removed:
            try:
                scene.remove_actors(o.get_actors(scene))
            except:
                pass
        
        for optic in opticList:
            optic.on_trait_change(self.trace_all, "update")
            optic.on_trait_change(self.render_vtk, "render")
        for optic in removed:
            optic.on_trait_change(self.trace_all, "update", remove=True)
            optic.on_trait_change(self.render_vtk, "render", remove=True)
        self.trace_all()
    
    def _rays_changed(self, rayList):
        scene = self.scene
        sources = [o.pipeline for o in rayList]
        mappers = [tvtk.PolyDataMapper(input_connection=s.output_port) for s in sources]
        actors = [tvtk.Actor(mapper=m) for m in mappers]
        for actor in actors:
            property = actor.property
            property.color = (1,0.5,0)
        scene.add_actors(actors)
        self.trace_all()
        
    def _probes_changed(self, probeList):
        scene = self.scene
        #del scene.actor_list[:]    
        for p in probeList:
            scene.add_actors(p.get_actors(scene))
        for probe in probeList:
            probe.on_trait_change(self.update_probes, "update")
            probe.on_trait_change(self.render_vtk, "render")
        self.trace_all()
        
    def _constraints_changed(self, constraintsList):
        for constraint in constraintsList:
            constraint.on_trait_change(self.trace_all, "update")
        self.trace_all()
        
    def _results_changed(self, resultsList):
        pass #not yet sure what we need to do here
        
    def update_probes(self):
        if self.scene is not None:
            self.render_vtk()
        
    def trace_all(self):
        if self._hold_off:
            self._update_requested=True
            return
        if not self._updating:
            self._updating = True
            self.update = True
            
    @contextmanager
    def hold_off(self):
        """A provides a context where the tracing operations are blocked so
        the user can edit parameters of the model without triggering multiple traces.
        A final trace occurs on exiting the context if required."""
        self._hold_off = True
        try:
            yield
        except:
            self._update_requested = False
            raise
        finally:
            self._hold_off = False
        if self._update_requested:
            self._update_requested = False
            self.trace_all()
        
    @on_trait_change("update", dispatch="queued")
    def do_update(self):
        optics = self.optics
        print( "trace") 
        next(counter)
        try:
            if optics is not None:
                self.prepare_to_trace()
                for o in optics:
                    o.intersections = []
                for ray_source in self.sources:
                    self.trace_ray_source(ray_source, optics)
                for probe in self.probes:
                    try:
                        probe.evaluate(self.sources)
                    except:
                        traceback.print_exc()
                for o in optics:
                    o.update_complete()
                for r in self.results:
                    try:
                        r.calc_result(self)
                    except:
                        traceback.print_exc()
            self.render_vtk()
            self._updating = False
        except:
            traceback.print_exc()
        
    def trace_detail(self, _async=False):
        optics = [o.clone_traits() for o in self.optics]
        for child, parent in zip(optics, self.optics):
            child.shadow_parent = parent
        sources = [s.clone_traits() for s in self.sources]
        for child, parent in zip(sources, self.sources):
            child.shadow_parent = parent
        probes = [p.clone_traits() for p in self.probes]
        for child, parent in zip(probes, self.probes):
            child.shadow_parent = parent
        if _async:
            self.thd = threading.Thread(target=self.async_trace, 
                                args=(optics, sources, probes))
            self.thd.start()
        else:
            self.async_trace(optics, sources, probes)
        
    def async_trace(self, optics, sources, probes):
        """called in a thread to do background tracing"""
        for o in optics:
            o.intersections = []
        for ray_source in sources:
            self.trace_ray_source_detail(ray_source, optics)
        for probe in probes:
            probe.find_intersections(ray_source)
        for o in optics:
            o.update_complete()
            
        #wx.CallAfter(self.on_trace_complete, optics, sources)
        FutureCall(0, self.on_trace_complete, optics, sources)
        
    def on_trace_complete(self, optics, sources):
        for s in sources:
            s.shadow_parent.copy_traits(s)
        for o in optics:
            o.shadow_parent.copy_traits(o)
        print("async trace complete")
        
    def render_vtk(self):
        if self.scene is not None:
            self.scene.render()
            
    def prepare_to_trace(self):
        """Called before a tracing operation is performed, to do
        all synchronisation between optics and their faces
        """
        face_sets = [o.faces for o in self.optics]
        for fs in face_sets:
            fs.sync_transforms()
            
        self.face_sets = face_sets
        
    def trace_ray_source(self, ray_source, optics):
        """trace a ray source asequentially, using the ctracer framework"""
        max_length = ray_source.max_ray_len
        rays = ray_source.input_rays #FIXME
        face_lists = self.face_sets
        rays.wavelengths = numpy.ascontiguousarray(ray_source.wavelength_list, numpy.double)
        try:
            traced_rays, all_faces = trace_rays(rays, face_lists, 
                                                recursion_limit=self.recursion_limit, 
                                                max_length=max_length)
            self.all_faces = all_faces
            ray_source.traced_rays = traced_rays
        finally:
            ray_source.data_source.modified()
        
    def trace_sequence(self, input_rays, faces_sequence):
        """
        Perform a sequential ray-trace. *** THIS IS BROKEN ***
        
        @param input_rays: a RayCollection instance
        @param optics_sequence: a list of Face instances or lists of Faces
        
        returns - the traced rays, as a list of RayCollections including
                the initial input rays
        """
        traced_rays = [rays]
        rays = input_rays
        for faces in faces_sequence:
            if isinstance(faces, Face):
                intersections = face.trace_rays(rays)
                mask = intersections['length']!=numpy.Infinity
                intersections = intersections[mask]
                points = intersections['point']
                children = face.eval_children(rays, points)
            else:
                intersections = numpy.column_stack([f.trace_rays(rays) for f in faces])
                shortest = numpy.argmin(intersections['length'], axis=1)
                ar = numpy.arange(size)
                lengths = intersections['length'][ar,shortest]
                
                #now remove infinite rays
                mask = lengths!=numpy.Infinity
                shortest = shortest[mask]
                ar = ar[mask]
                
            rays = children
            traces_rays.append(rays)
        return traced_rays
    
    def _save_btn_changed(self):
        dlg = FileDialog(action="save as",
                         wildcard=FileDialog.create_wildcard("Scene model", 
                                                             ["*.stp",
                                                              "*.step",
                                                              "*.wrl",
                                                              "*.vrml"]))
        ret = dlg.open()
        if ret != FD_OK: return
        filename = dlg.path
        fmap = {".stp": self.write_to_STEP,
                ".step": self.write_to_STEP,
                ".wrl": self.write_to_VRML,
                ".vrml": self.write_to_VRML}
        ext = os.path.splitext(filename)[-1].lower()
        try:
            fmap[ext](filename)
        except KeyError:
            self.write_to_STEP(filename)
    
    def write_to_VRML(self, fname):
        scene = self.scene
        if scene is not None:
            renwin = scene._renwin
            if filename:
                writer = tvtk.VRMLExporter(file_name=fname,
                                           render_window=renwin)
                writer.update()
                writer.write()
                
    def write_to_STEP(self, fname):
        from raypier.step_export import export_shapes2 as export_shapes
        optics = self.optics
        sources = self.sources
        shapes_colors = [_f for _f in (o.make_step_shape() for o in optics) if _f]
        shapes_colors.extend([_f for _f in [s.make_step_shape() for s in sources] if _f])
        
        shapes = [s for s,c in shapes_colors]
        colors = [c for s,c in shapes_colors]
        export_shapes(shapes, fname, colorList=colors)
        
    def ipython_view(self, width, height, view={}):
        import ipywidgets as widgets
        from IPython.display import Image, display, clear_output
        
        renderer = tvtk.Renderer()
        for actor in self.scene.actor_list:
            renderer.add_actor(actor)
        renderer.background = (1,1,0.8)
#         
        renderer.reset_camera()
        camera = renderer.active_camera
        if "position" in view:
            camera.position = view['position']
        if "focal_point" in view:
            camera.focal_point = view['focal_point']
        if "view_up" in view:
            camera.view_up = view['view_up']
#         
        renderWindow = tvtk.RenderWindow()
        renderWindow.off_screen_rendering = True
        renderWindow.add_renderer(renderer)
        renderWindow.size = (width, height)
        renderWindow.render()
        
        windowToImageFilter = tvtk.WindowToImageFilter()
        windowToImageFilter.input = renderWindow
        windowToImageFilter.update()
#          
        filename = "/dev/shm/temp_vtk_put.png"
        writer = tvtk.PNGWriter()
        writer.file_name = filename
        writer.write_to_memory = False
        writer.input_connection = windowToImageFilter.output_port
        writer.write()
        
        view_out = {}
        
        def show():       
            clear_output(wait=True)
            renderer.reset_camera_clipping_range()
            renderer.modified()
            renderWindow.render()
            windowToImageFilter.input = renderWindow
            windowToImageFilter.modified()
            windowToImageFilter.update()
            writer.write()
            view_out.update({"position": tuple(camera.position), 
                             "view_up": tuple(camera.view_up),
                             "focal_point": tuple(camera.focal_point)})
            return display(Image(filename), grp)
        
        def r_up(arg):
            camera.orthogonalize_view_up()
            camera.elevation(10)
            return show()
        
        def r_down(arg):
            camera.orthogonalize_view_up()
            camera.elevation(-10)
            return show()
        
        def r_left(arg):
            camera.orthogonalize_view_up()
            camera.azimuth(10)
            return show()
        
        def r_right(arg):
            camera.orthogonalize_view_up()
            camera.azimuth(-10)
            return show()
        
        def pan_left(arg):
            camera.orthogonalize_view_up()
            camera.yaw(-2)
            return show()
        
        def pan_right(arg):
            camera.orthogonalize_view_up()
            camera.yaw(2)
            return show()
        
        def pan_up(arg):
            camera.orthogonalize_view_up()
            camera.pitch(-2)
            return show()
        
        def pan_down(arg):
            camera.orthogonalize_view_up()
            camera.pitch(2)
            return show()
        
        def roll_left(arg):
            camera.roll(10)
            return show()
        
        def roll_right(arg):
            camera.roll(-10)
            return show()
        
        def zoom_in(arg):
            camera.dolly(1.2)
            return show()
            
        def zoom_out(arg):
            camera.dolly(0.8)
            return show()
        
        
        b1 = widgets.Button(description = '\u2191')
        b1.on_click(r_up)
        b2 = widgets.Button(description = '\u2193')
        b2.on_click(r_down)
        b3 = widgets.Button(description = '\u2190')
        b3.on_click(r_left)
        b4 = widgets.Button(description = '\u2192')
        b4.on_click(r_right)
        b5 = widgets.Button(description = '\u21ba')
        b5.on_click(roll_left)
        b6 = widgets.Button(description = '\u21bb')
        b6.on_click(roll_right)
        
        b7 = widgets.Button(description = '+')
        b7.on_click(zoom_in)
        b8 = widgets.Button(description = '-')
        b8.on_click(zoom_out)
        b9 = widgets.Button(description = '\u2190')
        b9.on_click(pan_left)
        b10 = widgets.Button(description = '\u2192')
        b10.on_click(pan_right)
        b11 = widgets.Button(description = '\u2191')
        b11.on_click(pan_up)
        b12 = widgets.Button(description = '\u2193')
        b12.on_click(pan_down)
        
        grp1 = widgets.HBox(border_style="solid",
                            border_width=1,
                            border_radius=5,
                            padding=2,
                            margin=1)
        grp1.children=[b1,b2,b3,b4,b5,b6]
        
        grp2 = widgets.HBox(border_style="solid",
                            border_width=1,
                            border_radius=5,
                            padding=2,
                            margin=1)
        grp2.children=[b7,b8,b9,b10,b11,b12]
        
        
        grp = widgets.HBox(children=[grp1,grp2])
        
        show()
        return view_out
        
        
    def render_bitmap(self, width, height, filename=None,
                      azimuth=15.0, elevation=30.0, roll=0.0,
                      zoom=1.0, pan_h=0.0, pan_v=0.0
                      ):
        renderer = tvtk.Renderer()
        for actor in self.scene.actor_list:
            renderer.add_actor(actor)
        renderer.background = (1,1,0.8)
        
        renderer.reset_camera()
        camera = renderer.active_camera
        camera.roll(roll)
        camera.elevation(elevation)
        camera.azimuth(azimuth)
        camera.dolly(zoom)
        camera.yaw(pan_h)
        camera.pitch(pan_v)
        
        renderWindow = tvtk.RenderWindow()
        renderWindow.off_screen_rendering = True
        renderWindow.add_renderer(renderer)
        renderWindow.size = (width, height)
        renderWindow.render()
         
        windowToImageFilter = tvtk.WindowToImageFilter()
        windowToImageFilter.input = renderWindow
        windowToImageFilter.update()
         
        writer = tvtk.PNGWriter()
        if filename is not None:
            writer.file_name = filename
            writer.write_to_memory = False
        else:
            writer.write_to_memory = True
        writer.input_connection = windowToImageFilter.output_port
        writer.write()
        #data = numpy.asarray(writer.result).tostring()
        return writer.result
        
    @on_trait_change("sources[]")
    def on_sources_changed(self, obj, name, removed, source_list):
        scene = self.scene
        for source in source_list:
            for actor in source.actors:
                scene.add_actor(actor)
            source.on_trait_change(self.trace_all, "update")
            source.on_trait_change(self.render_vtk, "render")
            
        for source in removed:
            scene.remove_actors(source.actors)
            source.on_trait_change(self.trace_all, "update", remove=True)
            source.on_trait_change(self.render_vtk, "render", remove=True)
        self.trace_all()
    
    
#use a singleton handler
controller = RayTraceModelHandler()
    
        
def on_dclick(*obj):
    print("objects", obj)
    obj[0].edit_traits(kind="live", parent=controller.info.ui.control)
    
    
        
tree_editor = TreeEditor(
                nodes=[
                       TreeNode(
                        node_for=[RayTraceModel],
                        children='',
                        auto_open=True,
                        label="=My Model",
                        view = View()
                        ),
                       TreeNode(
                        node_for=[RayTraceModel],
                        children='optics',
                        auto_open=True,
                        label="=Components",
                        view = View(),
                        ),
                       TreeNode(
                        node_for=[RayTraceModel],
                        children='sources',
                        auto_open=True,
                        label="=Ray Sources",
                        view = View(),
                        ),
                        TreeNode(
                        node_for=[RayTraceModel],
                        children='probes',
                        auto_open=True,
                        label="=Probes",
                        view = View(),
                        ),
                       TreeNode(
                        node_for=[Traceable],
                        children='',
                        auto_open=False,
                        label="name",
                        ),
                       TreeNode(
                        node_for=[BaseRaySource],
                        children='',
                        auto_open=True,
                        label="name",
                        ),
                       TreeNode(
                        node_for=[Probe],
                        children='',
                        auto_open=True,
                        label="name",
                        ),
                       TreeNode(
                        node_for=[RayTraceModel],
                        children='constraints',
                        auto_open=True,
                        label="=Constraints",
                        view = View()
                        ),
                       TreeNode(
                        node_for=[RayTraceModel],
                        children='results',
                        auto_open=True,
                        label="=Results",
                        view = View()
                        ),
                       TreeNode(
                        node_for=[BaseConstraint],
                        children='',
                        auto_open=True,
                        label="name",
                        ),
                       TreeNode(
                        node_for=[Face],
                        children='',
                        auto_open=False,
                        label="name",
                        ),
                       TreeNode(
                        node_for=[Result],
                        children='',
                        auto_open=False,
                        label="name",
                        ),
                       ],
                orientation='vertical',
                hide_root=True,
                on_dclick=on_dclick,
                )
    
ray_tracer_view = View(
                   HSplit(
                    VSplit(
                       Item('scene',editor=SceneEditor(),
                            height=600),
                       #Item('optics@', editor=ListEditor(use_notebook=True),
                       #     width=200),
                       Tabbed(
                           Item('ShellObj', editor=ShellEditor(share=False)),
                           Item('constraints', style="custom", editor=ListEditor(use_notebook=True)),
                           show_labels=False
                           ),
                       show_labels=False,
                       dock="vertical"
                       ),
                       VGroup(Item('Self', 
                                id="TracerModelID",
                                editor=tree_editor, width=200),
                            Item('save_btn'),
                            show_labels=False
                            ),
                    show_labels=False,
                    dock="horizontal",
                    id="raypier.model"
                   ),
                   resizable=True,
                   #height=500,
                   width=800,
                   id="raypier.view",
                   handler=controller,
                   menubar=menubar,
                   )
    
RayTraceModel.class_trait_view("traits_view", ray_tracer_view)
    
        




from raypier.tracer import RayTraceModel
from raypier.ellipsoids import Ellipsoid
from raypier.sources import ParallelRaySource, ConfocalRaySource

e = Ellipsoid(focus1=(-50,0,0),
                    focus2=(0,50,0),
                    Z_bounds=(-25,25),
                    size=100.)

#source = ParallelRaySource(origin=(-60,0,12),
#                           direction=(1,0,0),
#                           radius=10.)
source = ConfocalRaySource(focus=(-50,0,0),
                           direction=(1,0,0),
                           theta=15.,
                           working_dist=10.)

model = RayTraceModel(optics=[e,],
                      sources=[source,])

#model.update = True
#
#print source.traced_rays[0]

model.configure_traits()



"""
Functions for performing 2D phase unwrapping.
"""

import numpy


def unwrap2d(phase_array, anchor=(0,0)):
    """
    A basic phase unwrapper using numpy.unwrap and the Itoh method.
    
    phase_array - a (N,M) shaped array of values in the range -pi .. +pi
    """
    ###dPhi/dx
    ddx = numpy.diff(phase_array, axis=0) #shape=(N-1,M)
    ddy = numpy.diff(phase_array, axis=1) #shape=(N,M-1)
    
    ### Calc path-integral of phase delta over each 2x2 pixel cell
    residuals = ddx[:,:-1] - ddx[:,1:] - ddy[:-1,:] + ddy[1:,:]
    
    int1 = numpy.unwrap(phase_array, axis=0)
    int2 = numpy.unwrap(int1, axis=1)
    
    a,b = anchor
    offset = int((int2[a,b] - phase_array[a,b])/(2*numpy.pi))
    int2 -= offset*numpy.pi*2
    return int2, residuals
    
    
    
if __name__=="__main__":
    x_ = numpy.linspace(-0.1,0.1,100)
    
    x,y = numpy.meshgrid(x_,x_)
    z = 1 - numpy.sqrt(1 - x**2 - y**2)
    z *= 5000.
    
    noise = numpy.random.standard_normal(z.shape)*(numpy.pi/10)
    z += noise
    
    z_w = (z + numpy.pi)%(2*numpy.pi) - numpy.pi
    
    
    from matplotlib import pyplot as pp
    
    pp.imshow(z_w)
    pp.show()
    
    residuals, unwrapped = unwrap2d(z_w, anchor=(50,50))
    
    pp.imshow(residuals)
    pp.show()
    
    pp.imshow(unwrapped)
    pp.show()
    